#include "config.h"
#include <inttypes.h>
#include <Philips/arm-irq.h>
#include <Philips/VIC.hh>
#include <string.h>
#include <Util/rprintf.h> //;;

#include "I2C.hh"

//-----------------------------------------------------------------------------

const int      Retries = 3;		// I2C retries
const unsigned Timeout = 1000;		// I2C timeout, AKX ticks

I2Cbase *I2Cbase::objp[I2C_COUNT];

//-----------------------------------------------------------------------------

void
I2Cbase::init(int no)
{
    i2c = &I2C(no);
    i2c->conclr = I2CON_ENABLE;
    number = no;
    cmdlen = 0;
    datlen = 0;
    rate(100000);
}

void
I2Cbase::start(Vic::irq_t irqfcn)
{
    vic.vector(VIC_I2C(number), irqfcn);
    i2c->conset = I2CON_ENABLE;			// enable I2C
}

//-----------------------------------------------------------------------------

void
I2Cbase::enter()
{
  #ifdef AKX
    mutex.lock();
  #endif
}

void
I2Cbase::leave()
{
  #ifdef AKX
    mutex.unlock();
  #endif
}

void
I2Cbase::signal()
{
  #ifdef AKX
    event.signal();
  #endif
}

void
I2Cbase::reset()
{
  #ifdef AKX
    event.reset();
  #endif
}

void
I2Cbase::wait()
{
  #ifdef AKX
    event.wait(Timeout);
  #else
    while (state == I2STAT_NO_INFO)
	;
  #endif
}

//-----------------------------------------------------------------------------

void
I2Cbase::rate(unsigned rate)
{
    unsigned cntr = (PCLK + rate-1) / rate;	// round up
    if (cntr < 8)				// minimum allowed counter value is 4
	cntr = 8;
    i2c->clk_high = cntr - cntr / 2;		// high clock phase
    i2c->clk_low  = cntr / 2;			// low  clock phase
}

//-----------------------------------------------------------------------------

int
I2Cbase::transact()
{
    reset();

    state = I2STAT_NO_INFO;
    i2c->conclr = I2CON_STOP;
    i2c->conset = I2CON_ENABLE | I2CON_START | I2CON_ACK;

    wait();

    return state;
}

//-----------------------------------------------------------------------------

int
I2Cbase::write(unsigned addrByte,
	       const uint8_t *ctrlBuf, int ctrlLen,
	       const uint8_t *dataBuf, int dataLen)
{
    enter();
    
    addr   = addrByte;
    cmdbuf = ctrlBuf;
    cmdlen = ctrlLen;
    datbuf = (volatile uint8_t *)dataBuf;
    datlen = dataLen;
    int rc = transact();
    
    if (rc == I2STAT_MT_DATA_ACK)
	rc = Success;

    leave();
    return rc;
}

//-----------------------------------------------------------------------------

int
I2Cbase::read(unsigned addrByte,
	      const uint8_t *ctrlBuf, int ctrlLen,
	      uint8_t *dataBuf, int dataLen)
{
    int rc;
    enter();

    if (ctrlLen) {
	addr   = addrByte;
	cmdbuf = ctrlBuf;
	cmdlen = ctrlLen;
	datlen = 0;
	rc = transact();		// the WRITE transaction
    } else
	rc = I2STAT_MT_SLA_ACK;

    if (rc == I2STAT_MT_SLA_ACK) {
    	addr |= I2C_READ;
    	cmdlen = 0;
    	datbuf = dataBuf;
    	datlen = dataLen;
	rc = transact();		// the READ transaction

	if (state == I2STAT_MR_DATA_NACK && index == datlen)
	    rc = Success;
    }

    leave();
    return rc;
}

//-----------------------------------------------------------------------------

void
I2Cbase::interrupt()
{
    vic.nesting();

    // The non-vectored interrupt handler polls all possible interrupts and also 
    // catches spurious interrupts, so just return when disabled.
    if (!vic.enabled(VIC_I2C(number)) || !(i2c->con & I2CON_ENABLE))
	return;

    if (!(i2c->con & I2CON_IFLAG))
	return;

    int mystate = i2c->state;
    //+ puts("i2c:"); putu(mystate, 0, 16); putc('\n');

    switch (mystate >> 3) {

    case I2STAT_START >> 3:
	// 08 = start condition transmitted
    case I2STAT_REP_START >> 3:
	// 10 = repeated start condition transmitted
	index = 0;						// initialize buffer pointer
	//+ puts("addr:"); putu(addr, 0, 16); putc('\n');
	i2c->data = addr;					// transmit address byte
	i2c->conclr = I2CON_START;				// reset start condition
	break;
		
    case I2STAT_MT_SLA_ACK >> 3:
	// 18 = master transmit, slave address transmitted, ack received
    case I2STAT_MT_DATA_ACK >> 3:
	// 28 = master transmit, data byte transmitted, ack received
	if (index < cmdlen) {
	    //+ puts("tcmd:"); putu(index); putc(','); putu(cmdbuf[index],2,16); putc('\n');
	    i2c->data = cmdbuf[index];				// send next command byte
	    ++index;
	} else if (index-cmdlen < datlen) {
	    //+ puts("tdat:"); putu(index-cmdlen); putc(','); putu(datbuf[index-cmdlen],2,16); putc('\n');
	    i2c->data = datbuf[index-cmdlen];			// send next data byte
	    ++index;
	} else {
	    goto stop;
	}
	break;

    case I2STAT_MR_DATA_ACK >> 3:
	// 50 = master receive, data byte received, ack returned
    	//+ puts("rdat:");
	if (index < datlen) {
	    datbuf[index++] = i2c->data;			// store received byte
	    //+ putu(datbuf[index-1],2,16);
	}
    	//+ putc('\n');
    case I2STAT_MR_SLA_ACK >> 3:
	// 40 = master receive, slave address transmitted, ack received
	if (index >= datlen-1)
	    i2c->conclr = I2CON_ACK;				// nack to last byte
	break;

    case I2STAT_MR_DATA_NACK >> 3:
	// 58 = master receive, data byte received, nack returned
    	//+ puts("rdat:");
	if (index < datlen) {
	    datbuf[index++] = i2c->data;			// store received byte
	    //+ putu(datbuf[index-1],2,16);
	}
    	//+ putc('\n');
	goto stop;

    case I2STAT_ARB_LOST >> 3:
	// 38 = arbitration lost
	i2c->conset = I2CON_START;				// set start condition again
	break;

    case I2STAT_MT_SLA_NACK >> 3:
	// 20 = master transmit, slave address transmitted, nack received
    case I2STAT_MT_DATA_NACK >> 3:
	// 30 = master transmit, data byte transmitted, nack received
    case I2STAT_MR_SLA_NACK >> 3:
	// 48 = master receive, slave address transmitted, nack received
    stop:
	//+ puts("stop\n");
	i2c->conset = I2CON_STOP;				// set stop condition
    default:
	state = mystate;
	signal();
	break;
    }

    i2c->conclr = I2CON_IFLAG;					// clear interrupt flag, release SCK
}
