#ifndef I2C_HH_
#define I2C_HH_

//-----------------------------------------------------------------------------

//
// Requires I/O definitions from lpc2xxx.h and
//	PCLK		Peripheral clock frequency.
//	AKX			Use RTOS.
//

#include "config.h"

#ifdef AKX
  #include <Mutex.h>
  #include <Sema.h>
#endif

//-----------------------------------------------------------------------------

class I2Cbase {
public:

    // values which are not used by I2C state machine
    enum {  Success = 0x01 };	// ok

    // set transmission rate
    void rate (unsigned);

    // write to I2C device
    // written byte stream is <addr,ctrl,data>
    // return Success on success, or I2C status
    int	write (unsigned addr,
	       const uint8_t *ctrlBuf, int ctrlLen,
	       const uint8_t *dataBuf, int dataLen);

    // read from I2C device
    // written byte stream is <addr,ctrl>
    // data returned is in <data>
    // return Success on success, or I2C status
    int read (unsigned addr,
	      const uint8_t *ctrlBuf, int ctrlLen,
	      uint8_t *dataBuf, int dataLen);

    void start (Vic::irq_t irqfcn);

    void interrupt ();

protected:
    static I2Cbase *objp[I2C_COUNT];	// I2C index to object

    I2Cbase (int no) { init(no); }
    void init (int no);
    
    int transact ();

    void inline enter ();
    void inline leave ();
    void inline signal ();
    void inline reset ();
    void inline wait ();

private:
  #ifdef AKX
	OS::Mutex				mutex;
	OS::Sema				event;
  #endif
    volatile I2C_t *		i2c;
    volatile const uint8_t *	cmdbuf;
    volatile uint8_t *		datbuf;
    volatile uint8_t		state;			// I2C state machine
    volatile uint16_t		index;			// walks thru cmdbuf then datbuf
    volatile uint16_t		cmdlen, datlen;
             uint8_t		addr, number;
};

//-----------------------------------------------------------------------------

template<int no>
class I2Cc : public I2Cbase {
public:
    I2Cc () : I2Cbase(no) {
    	I2Cbase::objp[no] = this;
    }
    void start () {
	I2Cbase::start(irq);
    }
protected:
    static IRQ_FUNCTION void irq();
};

template<int no>
void I2Cc<no>::irq()
{
    I2Cbase::objp[no]->interrupt();
}

//-----------------------------------------------------------------------------
#endif // I2C_HH_
